import http.client


def shutdown(port: int = 8666) -> str:
    connection = http.client.HTTPConnection("localhost", port=port)
    connection.request("GET", "/__SHUTDOWN")
    result = connection.getresponse().fp.read().decode()
    connection.close()
    return result



import pathlib
import re
from hakisto import Logger
from pykeepass import PyKeePass
from pykeepass.entry import reserved_keys

logger = Logger("imuthes.keepass")

reserved_keys = set([i.lower() for i in reserved_keys if i not in ("History", "IconID", "Times", "otp")])

URL_RE = re.compile(r"\{.+?}")


class KeePass(PyKeePass):
    def __init__(
        self,
        filename: pathlib.Path,
        password: str,
        keyfile: pathlib.Path = None,
    ):
        super().__init__(filename=filename, password=password, keyfile=keyfile)

    def get(self, path: str | list[str]) -> dict[str, str]:
        """Get Entry as Dict"""
        logger.verbose(f"getting {path}")
        if isinstance(path, str):
            path = path.strip("/").split("/")
        entry = self.find_entries(path=path)
        if not entry:
            raise KeyError("/".join(path))
        result = {k: getattr(entry, k) for k in reserved_keys}
        result.update(entry.custom_properties)
        for k in result:  # resolve internal refs
            if result[k] and result[k].startswith("{REF:"):
                result[k] = entry.deref(k)
        if result["url"]:
            for m in URL_RE.finditer(result["url"]):
                var = m.group()[3:-1] if m.group().startswith("{S:") else m.group()[1:-1].lower()
                if var in result:
                    result["url"] = result["url"].replace(m.group(), result[var])
        return result

    # noinspection PyShadowingBuiltins
    def get_property(self, path: str | list[str], property: str) -> str:
        """Get Property as String"""
        logger.verbose(f'Getting property "{property}" from "{path}"')
        entry = self.get(path=path)
        return entry[property]



import pathlib
import click
from .server import server as _server
from .get import get as _get
from .shutdown import shutdown as _shutdown
from hakisto.click import (
    hakisto_inline_location,
    hakisto_severity,
    hakisto_short_trace,
    hakisto_process_all,
    hakisto_use_color,
)


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@hakisto_severity()
@hakisto_use_color()
@hakisto_short_trace()
@hakisto_inline_location()
@click.pass_context
def main(ctx, **kwargs):
    """Keepass utilities"""
    hakisto_process_all(**kwargs)
    ctx.ensure_object(dict)


@main.group()
@click.option("-p", "--port", type=int, default=8666, help="Port (default: 8666)")
@click.pass_context
def server(ctx, port):
    """Rest server"""
    ctx.obj["port"] = port


@server.command()
@click.option("-p", "--passphrase", prompt=True, hide_input=True, help="Passphrase")
@click.option(
    "-k",
    "--keyfile",
    type=click.Path(exists=True, dir_okay=False, readable=True, allow_dash=False, path_type=pathlib.Path),
    help="Keyfile",
)
@click.argument(
    "filename", type=click.Path(exists=True, dir_okay=False, readable=True, allow_dash=False, path_type=pathlib.Path)
)
@click.pass_context
def start(ctx, passphrase, filename, keyfile):
    """Start Server at PORT listening to 127.0.0.1 (localhost)."""
    _start(filename=filename, passphrase=passphrase, keyfile=keyfile, port=ctx.obj["port"])


def _start(filename: pathlib.Path, passphrase: str, keyfile: pathlib.Path, port: int = 8666):
    """Start Server"""
    _server(filename=filename, password=passphrase, keyfile=keyfile, port=port)


@server.command()
@click.option(
    "-s",
    "--secret",
    type=click.Path(exists=True, dir_okay=False, readable=True, allow_dash=False, path_type=pathlib.Path),
    help="File with encrypted secret",
)
@click.option(
    "-k",
    "--keyfile",
    type=click.Path(exists=True, dir_okay=False, readable=True, allow_dash=False, path_type=pathlib.Path),
    help="Keyfile",
)
@click.argument(
    "filename", type=click.Path(exists=True, dir_okay=False, readable=True, allow_dash=False, path_type=pathlib.Path)
)
@click.pass_context
def start_secret(ctx, secret, filename, keyfile):
    """Start Server at PORT listening to 127.0.0.1 (localhost) using SECRET file."""
    from ..crypt import read

    passphrase = read(secret)
    _start(filename=filename, passphrase=passphrase, keyfile=keyfile, port=ctx.obj["port"])


@server.command()
@click.pass_context
def shutdown(ctx):
    """Shutdown server at PORT"""
    print(_shutdown(port=ctx.obj["port"]))


@server.command()
@click.argument("path")
@click.argument("property")
@click.pass_context
def get(ctx, path, property):
    """
    Retrieve PATH from local REST server and print content of PROPERTY.
    """
    print(_get(path=path, property=property, port=ctx.obj["port"]))



# ------------------------------------------------------------------------------
#  Imuthes by NetLink Consulting GmbH
#
#  Copyright (c) 2025. Bernhard W. Radermacher
#
#  This program is free software: you can redistribute it and/or modify it under
#  the terms of the GNU Lesser General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
#  details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

__import__("pkg_resources").declare_namespace(__name__)  # pragma: no cover



import http.client
import json


# noinspection PyShadowingBuiltins
def get(path: str, property: str, port: int = 8666) -> str:
    path = path.strip("/")
    connection = http.client.HTTPConnection("localhost", port=port)
    connection.request("GET", f"/{path}")
    j = json.loads(connection.getresponse().fp.read().decode())
    connection.close()
    result = j.get(property) or ""
    result = "\n".join(result.splitlines())  # need to normalize newline
    return result



# Localhost REST server for Keepass

import os
import pathlib
import signal

import fastapi
import uvicorn

from .keepass import KeePass

app = fastapi.FastAPI()


# noinspection PyUnresolvedReferences
@app.get("/{full_path:path}")
async def get_value(full_path: str, p: str = None):
    if full_path.startswith("__SHUTDOWN"):
        os.kill(os.getpid(), signal.SIGTERM)
        return fastapi.Response(status_code=200, content="Server shutting down...")
    try:
        if p:
            return app.state.kp.get_property(full_path, p)
        return app.state.kp.get(full_path)
    except KeyError:
        raise fastapi.HTTPException(status_code=404, detail=f"{full_path} not found")


# noinspection PyUnresolvedReferences
def server(
    filename: pathlib.Path,
    password: str,
    keyfile: pathlib.Path = None,
    port: int = 8666,
) -> None:
    """Start REST server to read KeePass

    :param filename: KeePass Database file
    :param password: Password for KeePass database
    :param keyfile:
    :param port: Server port
    """
    app.state.kp = KeePass(filename=filename, password=password, keyfile=keyfile)
    uvicorn.run(app, port=port, log_level="info")
